/*
 * Copyright (C) 1997-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */ 


#include "server.h"

extern struct misc_meta_data *mmd;
extern struct gengetopt_args_info conf;
extern struct dbtool dblist[];

static int com_si(int, int, char **);
static int com_si(int, int, char **);
static int com_version(int, int, char **);
static int com_sb(int, int, char **);
static int com_sc(int, int, char **);
static int com_stat(int, int, char **);
static int com_help(int, int, char **);
static int com_hup(int, int, char **);
static int com_term(int, int, char **);
static int com_play(int, int, char **);
static int com_stop(int, int, char **);
static int com_pause(int, int, char **);
static int com_next(int, int, char **);
static int com_nomore(int, int, char **);
static int com_cdt(int, int, char **);
static int com_ff(int, int, char **);
static int com_jmp(int, int, char **);


/* commands that are handled by the server itself */
static struct command cmd_struct[] = {
	{
		.name = "cdt",
		.handler = com_cdt,
		.perms = DB_READ | DB_WRITE,
		.description = "change database tool",
		.usage = "cdt name_of_new_dbtool",
		.help =	
                "Deactivate current dbtool and activate \"name_of_new_dbtool\" \n"
	},
	{
		.name = "ff",
		.handler = com_ff,
		.perms = AFS_READ | AFS_WRITE,
		.description = "jmp amount of time forwards or backwards "
			"in current song",
		.usage = "ff n[-]",
		.help =	
		"Jump n seconds forward or backward.\n"
	},
	{	
		.name = "help",	
		.handler = com_help,
		.perms = 0,
		.description = "print help text",
		.usage = "help [command]",
		.help =
                "Without any arguments, help prints a list of        \n"
                "availible commands. When issued with a command name \n"
                "as first argument, help prints out a description for\n"
                "that command.                                       \n"
	},
	{	
		.name = "hup",
		.handler = com_hup,
		.perms = AFS_WRITE,
		.description = "force reload of config file and log file",
		.usage = "hup",
		.help =
                "After rereading the config file, a signal is sent to\n"
                "all children which forces them to close/reopen the  \n"
                "log file.                                           \n"
	},
	{
		.name = "jmp",
		.handler = com_jmp,
		.perms = AFS_READ | AFS_WRITE,
		.description = "jmp to given position in current song",
		.usage = "jmp [n]",
		.help =
		"Jump to position n, where 0 <= n <= 100\n"
	},
	{
		.name = "next",
		.handler = com_next,
		.perms = AFS_READ | AFS_WRITE,
		.description = "skip rest of current song",
		.usage = "next",
		.help =	
		"If stopped or paused, start playing.\n"
	},
	{
		.name = "nomore",
		.handler = com_nomore,
		.perms = AFS_READ | AFS_WRITE,
		.description = "stop playing after current song is over",
		.usage = "nomore", 
		.help = 
		"Use nomore instead of stop if you don't like\n"
		"sudden endings.\n"
	},
	{
		.name ="pause",
		.handler = com_pause,
		.perms = AFS_READ | AFS_WRITE,
		.description = "pause current song",
		.usage = "pause",
		.help =	
		"Use play to restart.\n"
	},
	{
		.name = "play",	
		.handler = com_play,
		.perms = AFS_READ | AFS_WRITE,
		.description = "start playing or resume playing when paused",
		.usage = "play [stream]",
		.help = 
		"Start/continue streaming.\n"
	},
	{
		.name = "sb",
		.handler = com_sb,
		.perms = AFS_READ,
		.description = "print status bar for current song",
		.usage = "sb [n]",
		.help =
                "Without any arguments, sb continuously prints a     \n"
                "status bar of the form                              \n"

                        "\n\t\"12:34 [56:12] (56%) filename\"\n\n"

                "indicating playing time, remaining time, percentage \n"
                "and the name of the file beeing streamed.  Use the  \n"
                "optional number n to let stat exit after having     \n"
                "displayed the status bar n times.                   \n"
	},
	{
		.name = "sc",
		.handler = com_sc,
		.perms = AFS_READ,
		.description = "print name of song whenever song changes",
		.usage = "sc",
		.help =
                "sc prints exactly one line (the filename of the song being "
		"played) whenever a song change was detected.\n"
	},
	{
		.name = "si",
		.handler = com_si,
		.perms = 0,
		.description = "print server info",
		.usage = "si",
		.help =
                "Print server uptime and other information.      \n"
	},
	{
		.name = "stat",	
		.handler = com_stat,
		.perms = AFS_READ, 
		.description = "print status info for current song",
		.usage = "stat [n]",
		.help =
                "Without any arguments, stat continuously prints     \n"
                "status messages of the song being streamed. Use the \n"
                "optional number n to let stat exit after having     \n"
                "displayed status n times.                           \n"
	},
	{
		.name = "stop",	
		.handler = com_stop,
		.perms = AFS_READ | AFS_WRITE,
		.description = "stop playing",
		.usage = "stop",
		.help =	
                "When stopped, server still accepts connections,     \n"
                "so you can still connect to the server and issue    \n"
                "commands, eg. play.                                 \n"
	},
	{
		.name = "term",	
		.handler = com_term,
		.perms = AFS_READ | AFS_WRITE,
		.description = "terminate server and audio file sender",
		.usage = "term",
		.help =
                "Shuts down the server. Instead of this command, you \n"
                "can also send SIGINT or SIGTERM. It should never be \n"
                "necessary to send SIGKILL.                          \n"
	},
	{
		.name = "version",
		.handler = com_version,
		.perms = 0,
		.description = "print server's version",
		.usage = "version",
		.help =	
		"Show version and other info\n"
	},
	/* this indicates the end of the list. Do not touch. */
	{ 
		.name = NULL,
	}
};

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mmd_lock();
	*new_mmd = *mmd;
	mmd_unlock();
}

/*
 * compute human readable string containing
 * afs_status for given integer value
 */
static char *afs_status_tohuman(unsigned int flags)
{
	if (flags & AFS_PLAYING)
		return para_strdup("playing");
	else if (flags & AFS_NEXT)
		return para_strdup("stopped");
	else
		return para_strdup("paused");
}

/*
 * construct string containing uptime
 */
static char *uptime_str(int i)
{
	return make_message("%i:%02i:%02i", i / 86400,
		(i / 3600) % 24, (i / 60) % 60);
}

/*
 * return human readable permission string. Never returns NULL.
 */
char *cmd_perms_itohuman(unsigned int perms) 
{
	char *msg = para_malloc(7 * sizeof(char));
	
	msg[0] = perms & DB_READ? 'd' : '-';
	msg[1] = perms & DB_WRITE? 'D' : '-';
	msg[2] = perms & AFS_READ? 'a' : '-';
	msg[3] = perms & AFS_WRITE? 'A' : '-';
	msg[4] = '\0';
	return msg;
}

/*
 * Never returns NULL.
 */
static char *afs_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = (flags & AFS_PLAYING)? 'P' : '_';
	msg[1] = (flags & AFS_NOMORE)? 'O' : '_';
	msg[2] = (flags & AFS_NEXT)? 'N' : '_';
	msg[3] = (flags & AFS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}


/* 
 * compute status bar string. Never returns NULL
 */
char *get_sb_string(struct misc_meta_data *nmmd)
{
	char *base;
	long long unsigned secs = 0, rsecs = 0;

	base = para_basename(nmmd->filename);
	if (!base || !base[0] || !nmmd->chunks_total)
		return para_strdup("");
	secs = (long long) nmmd->seconds_total * nmmd->chunks_sent
		/ nmmd->chunks_total;
	rsecs = (long long) nmmd->seconds_total * 
		(nmmd->chunks_total - nmmd->chunks_sent)
		/ nmmd->chunks_total; 
	return make_message("%llu:%02llu [%llu:%02llu] (%lu%%) %s",
		secs / 60, secs % 60, 
		rsecs / 60, rsecs % 60,
		100 * ((nmmd->chunks_sent + 5) / 10)
			/ ((nmmd->chunks_total + 5) / 10),
		base
	);
}

static char *compute_status_volatile(struct misc_meta_data *nmmd)
{
	char *bar, *ret;
	char *status, *flags; /* afs status info */
	char *ut = uptime_str(server_uptime(GET));

	status = afs_status_tohuman(nmmd->afs_status_flags);
	flags = afs_get_status_flags(nmmd->afs_status_flags);
	bar = get_sb_string(nmmd);

	ret = make_message(
		"status_bar:%s\n"
		"status:%s\n"
		"status_flags:%s\n"
		"dbtool:%s\n"
		"format:%s\n"
		"uptime:%s\n"
		"%s",
		bar,
		status,
		flags,
		dblist[nmmd->dbt_num].name,
		audio_format_name(nmmd->audio_format),
		ut,
		nmmd->audio_file_info
	);
	free(bar);
	free(flags);
	free(status);
	return ret;
}

static char *get_status(struct misc_meta_data *nmmd)
{
	return make_message("file_size:%lu\n"
		"mtime:%s"
		"length:%i\n"
		"length_min:%i\n"
		"length_sec:%02i\n"
		"num_played:%u\n",
		nmmd->size / 1024,
		ctime(&mmd->mtime),
		nmmd->seconds_total,
		nmmd->seconds_total / 60,
		nmmd->seconds_total % 60,
		nmmd->num_played);
}

/* server info */
static int com_si(int fd, int argc, char **argv)
{
	int i, ret;
	char *ut = uptime_str(server_uptime(GET));
	char *dbtools = NULL;

	mmd_lock();
	for (i = 0; dblist[i].name; i++) {
		dbtools = para_strcat(dbtools, dblist[i].name);
		dbtools = para_strcat(dbtools, " ");
	}
	ret = send_va_buffer(fd, "up: %s\nplayed: %u\n"
		"connections (total/accepted): %u/%u\n"
		"active connections: %u\n"
		"current loglevel: %i\n"
		"supported database tools: %s\n"
		"current database tool: %s\n"
		"supported audio formats: %s\n",
		ut, mmd->num_played, mmd->num_connects,
		mmd->num_commands, mmd->active_connections, 
		conf.loglevel_arg, 
		dbtools,
		dblist[mmd->dbt_num].name,
		afs_supported_audio_formats()
	);
	mmd_unlock();
	free(ut);
	if (dbtools)
		free(dbtools);
	return ret;
}

/* version */
static int com_version(int socket_fd, int argc, char **argv)
{
	if (send_buffer(socket_fd, "para_server version " VERSION  " | " 
			COPYRIGHT "\n" "built: " BUILD_DATE " (" SYSTEM ", " 
			CC_VERSION ")\n ") < 0)
		return -1;
	return 0;
}

/* sc */
static int com_sc(int socket_fd, int argc, char **argv)
{
	char *name;
	int ret, old = 0, update = 0;
repeat:
	mmd_lock();
	if (old != mmd->num_played) {
		old = mmd->num_played;
		update = 1;
		name = para_strdup(mmd->filename);
	}
	mmd_unlock();
	if (update) {
		name = para_strcat(name, "\n");
		ret = send_buffer_ll(socket_fd, name, DEBUG);
		free(name);
		if (ret < 0)
			return ret;
		update = 0;
	}
	usleep(500000);
	goto repeat;
}

/* sb */
static int com_sb(int socket_fd, int argc, char **argv)
{
	char *sb;
	int ret, nr = -1; 	/* status bar will be printed that many 
		 	 	 * times. Negative value means: print 
				 * forever
				 */
	if (argc)
		nr = atoi(argv[1]);
	while (nr) {
		mmd_lock();
		sb = get_sb_string(mmd);
		sb = para_strcat(sb, "\n");
		mmd_unlock();
		ret = send_buffer_ll(socket_fd, sb, DEBUG);
		free(sb);
		if (ret < 0)
			return ret;
		if (nr == 1)
			return 1;
		usleep(500000);
		if (nr > 0) 
			nr--;
	}
	return 1;
}

/* stat */
static int com_stat(int socket_fd, int argc, char **argv)
{
	char *old_stat = NULL, *old_dbinfo = NULL;
	int ret, num = 0;/* status will be printed that many 
			  * times. num <= 0 means: print forever
			  */
	int numplayed = 0; /* used to detect song change */
	struct misc_meta_data tmp, *nmmd = &tmp;

	if (argc)
		num = atoi(argv[1]);
	for (;;) {
		char *stat;

 		mmd_dup(nmmd);
		stat = compute_status_volatile(nmmd);

		if (!old_stat || num > 0 || strcmp(stat, old_stat)) {
			ret = send_buffer_ll(socket_fd, stat, INFO);
			free(old_stat);
			old_stat = stat;
			if (ret < 0) 
				goto out;
		} else
			free(stat);
		
		if (!old_dbinfo || num > 0 || strcmp(nmmd->dbinfo, 
						old_dbinfo)) {
			ret = send_buffer_ll(socket_fd, nmmd->dbinfo, INFO);
			if (old_dbinfo)
				free(old_dbinfo);
			old_dbinfo = para_strdup(nmmd->dbinfo);
			if (ret < 0) 
				goto out;
		}
		if (numplayed != nmmd->num_played) {
			char *s = get_status(nmmd);
			ret = send_buffer_ll(socket_fd, s, INFO);
			free(s);
			if (ret < 0)
				goto out;
			numplayed = nmmd->num_played;
		}
		ret = 1;
		if (num == 1) 
			goto out;
		usleep(500000);
	}
out:
	return ret;
}

static int send_description(int fd, struct command *cmd, int handler, int num)
{
	int ret, i;

	for (i = 1; cmd->name && (!num || i <= num); cmd++, i++) {
		char *perms = cmd_perms_itohuman(cmd->perms);
		ret = send_va_buffer(fd, "%s\t%s\t%s\t%s\n", cmd->name, 
			handler? "dbtool" : "server", 
			perms,
			cmd->description);
		free(perms);
		if (ret < 0)
			return ret;
	}
	return 1;
}

static struct command *get_cmd_ptr(char *name, int *handler)
{
	struct command *cmd = cmd_struct;
	
	if (handler)
		*handler = 0; /* server commands */
	for (cmd = cmd_struct; cmd->name; cmd++)
		if (!strcmp(cmd->name, name))
			return cmd;
	/* not found, look for dbtool commands */
	if (handler)
		*handler = 1; /* (1st) database tool */
	mmd_lock();
	cmd = dblist[mmd->dbt_num].cmd_list;
	mmd_unlock();
	for (; cmd->name; cmd++)
		if (!strcmp(cmd->name, name))
			return cmd;
	return NULL;
}

/* help */
static int com_help(int fd, int argc, char **argv)
{
	struct command *cmd;
	char *perms;
	int ret, handler;

	if (!argc) { 
		/* no argument given, print list of commands */
		if ((ret = send_description(fd, cmd_struct, 0, 0)) < 0)
			return ret;
		mmd_lock();
		cmd = dblist[mmd->dbt_num].cmd_list;
		mmd_unlock();
		return send_description(fd, cmd, 1, 0);
	}
	/* argument given for help */ 
	cmd = get_cmd_ptr(argv[1], &handler);
	if (!cmd)
		return send_buffer(fd, "Unknown command\n");
	perms = cmd_perms_itohuman(cmd->perms);
	ret = send_va_buffer(fd,
		"NAME:\n~~~~~\n\t%s -- %s\n\n"
		"USAGE:\n~~~~~~\n\t%s\n\n"
		"DESCRIPTION:\n~~~~~~~~~~~~\n%s\n"
		"HANDLER/MODE:\n~~~~~~~~~~~~~\n"
		"This command is handled by %s\n\n"
		"PERMS:\n~~~~~~\n"
		"Needed privileges for %s: %s\n", 
		argv[1],
		cmd->description,
		cmd->usage,
		cmd->help,
		handler? "dbtool" : "server",
		argv[1],
		perms
	);
	free(perms);
	return ret;
}

/* hup */
static int com_hup(int socket_fd, int argc, char **argv)
{
	kill(getppid(), SIGHUP);
	return 0;
}

/* term */
static int com_term(int socket_fd, int argc, char **argv)
{
	kill(getppid(), SIGTERM);
	return 0;
}

static int com_play(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags |= AFS_PLAYING;
	mmd_unlock();
	return 1;

}

/* stop */
static int com_stop(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags &= ~AFS_PLAYING;
	mmd->afs_status_flags &= ~AFS_REPOS;
	mmd->afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 0;
}

/* pause */
static int com_pause(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags &= ~AFS_PLAYING;
	mmd->afs_status_flags &= ~AFS_NEXT;
	mmd_unlock();
	return 0;
}

static int com_cdt(int fd, int argc, char **argv)
{
	int i, ret;

	if (!argc) {
		char *dbtool;
		mmd_lock();
		dbtool = para_strdup(dblist[mmd->dbt_num].name);
		mmd_unlock();
		ret = send_va_buffer(fd, "%s\n", dbtool);
		free(dbtool);
		return ret;
	}
	for (i = 0; dblist[i].name; i++) {
		if (strcmp(dblist[i].name, argv[1]))
			continue;
		mmd_lock();
		mmd->dbt_change = i;
		mmd_unlock();
		return 1;
	}
	return send_buffer(fd, "No such database tool");
}

/* next */
static int com_next(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 0;
}

/* nomore */
static int com_nomore(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags ^= AFS_NOMORE;
	mmd_unlock();
	return 0;
}
/* ff */
static int com_ff(int socket_fd, int argc, char **argv)
{
	int ret, sign = 1;
	char c;
	unsigned int i, skip;

	if (!argc)
		return -1;
	if (!(ret = sscanf(argv[1], "%u%c", &i, &c)))
		return -1;
	if (ret > 1 && c == '-') 
		sign = -1; /* jmp backwards */
	mmd_lock();
	if (!mmd->size)
		goto out;
	skip = i * ((mmd->chunks_total + 5) / 10)  / 
		((mmd->seconds_total + 5) / 10);
	if (sign == 1 && skip + mmd->chunks_sent > mmd->chunks_total) {
		mmd->afs_status_flags |= AFS_NEXT;
		goto out;
	}
	mmd->afs_status_flags |= AFS_REPOS;
	mmd->afs_status_flags &= ~AFS_NEXT;
	if (sign == -1 && mmd->chunks_sent < skip) {
		mmd->chunks_sent = 0;
		goto out;
	}
	mmd->chunks_sent += skip * sign;
out:
	mmd_unlock();
	return 0;
}

/* jmp */
static int com_jmp(int socket_fd, int argc, char **argv)
{
	long unsigned int i;

	if (!argc)
		return -1;
	if (sscanf(argv[1], "%lu", &i) <= 0)
		return -1;
	if (i > 100) 
		i = 100;
	mmd_lock();
	clog(DEBUG, "%s: jumping to %lu/%lu\n", __func__, 
		mmd->chunks_total / 100 * i, mmd->chunks_total);
	mmd->chunks_sent = mmd->chunks_total / 100 * i;
	mmd->afs_status_flags |= AFS_REPOS;
	mmd->afs_status_flags &= ~AFS_NEXT;
	mmd_unlock();
	return 0;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms. 
 * Returns 0 if perms are sufficient, -1 otherwise.
 */
int check_perms(unsigned int perms, struct command *cmd_ptr) 
{
	clog(DEBUG, "%s: checking permissions\n", __func__);
	return (cmd_ptr->perms & perms) < cmd_ptr->perms ? -1 : 0;
}

/*
 * Parse first string from *cmd and lookup in table of valid commands.
 * On error, NULL is returned. 
 */
struct command *parse_cmd(char *cmdstr) 
{
	char buf[255];
	struct command *cmd = NULL;
	int n = 0;

	sscanf(cmdstr, "%200s%n", buf, &n);
	if (!n)
	        return NULL;
	buf[n] = '\0';
	cmd = get_cmd_ptr(buf, NULL);
	if (cmd)
		clog(DEBUG, "%s: found command %s.\n", __func__, cmd->name);
	else 
		clog(ERROR, "%s: unknown command: %s\n", __func__, buf);
	return cmd;
}
